# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Column Classes

This module defines several specialized column types, aimed at providing efficient
memory usage and computational speed for various SQL queries. The column types differ
in their internal data representation but expose a common interface for manipulation
and query operations.

Features:
    - All column types (except FlatColumn & FunctionColumn) use compressed internal
      representations to conserve memory.
    - The compressed data is exposed through a 'values' property. This allows direct
      operations on compressed data, which usually involves fewer items compared to the
      uncompressed list.
    - Each column type provides a 'materialize' method to expand the compressed data
      into its uncompressed form, facilitating query operations that require a full
      column of data.

Column Types:
    - SparseColumn: Handles sparse data by only storing non-default values.
    - DictionaryColumn: Uses a dictionary to encode a finite set of string or
      numerical values.
    - RLEColumn: Utilizes Run-Length Encoding for sequences of repeating elements.
    - ConstantColumn: Represents a column of constant values using a single value and a
      length attribute.
    - FlatColumn: Standard column type that stores data in an uncompressed form.
    - FunctionColumn: Used when the column is the result of a function, useful as a
      placeholder column type.

By leveraging these compressed representations, Orso aims to provide a data store that
is both memory-efficient and computationally fast. For most operations, manipulating
the compressed data directly, bypassing the need to materialize the full column,
leading to faster evaluation.

Example:
    sparse_col = SparseColumn(name='col1', type=OrsoTypes.INTEGER, values=[1, None, 3])
    sparse_col.values *= 2  # Perform operation directly on compressed data
    full_array = sparse_col.materialize()  # Get the full, uncompressed array when needed

"""

from collections import defaultdict
from dataclasses import _MISSING_TYPE
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from dataclasses import fields
from decimal import Decimal
from enum import Enum
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import MutableMapping
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from warnings import warn

import numpy
import orjson
from data_expectations import Expectation

from orso.exceptions import ColumnDefinitionError
from orso.exceptions import DataValidationError
from orso.exceptions import ExcessColumnsInDataError
from orso.tools import arrow_type_map
from orso.tools import random_string
from orso.types import ORSO_TO_PYTHON_MAP
from orso.types import PYTHON_TO_ORSO_MAP
from orso.types import OrsoTypes

_MISSING_VALUE: str = str()


class ColumnDisposition(Enum):
    NAME = "name"
    AGE = "age"


class SchemaExpectation(Expectation):
    column = _MISSING_VALUE

    @classmethod
    def load(cls: Type["Expectation"], serialized: Union[Dict[str, Any], str]) -> "Expectation":
        """
        Loads a serialized Expectation and returns it as an instance.

        Parameters:
            serialized: Serialized Expectation as a dictionary or JSON string.

        Returns:
            An Expectation instance populated with the serialized data.
        """
        import json
        from copy import deepcopy

        if isinstance(serialized, str):
            serialized = dict(json.loads(serialized))
        serialized_copy: dict = deepcopy(serialized)
        if "expectation" not in serialized_copy:
            raise ValueError("Missing 'expectation' key in Expectation.")
        expectation = serialized_copy.pop("expectation")
        column = serialized_copy.pop("column", _MISSING_VALUE)
        ignore_nulls = serialized_copy.pop("ignore_nulls", True)
        config = serialized_copy
        return Expectation(
            expectation=expectation, column=column, ignore_nulls=ignore_nulls, config=config
        )


@dataclass(init=False)
class FlatColumn:
    """
    This is a standard column type.
    Unlike the other column types, we don't store the values for this Column
    here, we read them from the underlying data structure (orso/pyarrow/velox).
    """

    name: str
    default: Optional[Any] = None
    type: OrsoTypes = OrsoTypes._MISSING_TYPE
    description: Optional[str] = None
    disposition: Optional[ColumnDisposition] = None
    aliases: Optional[List[str]] = field(default_factory=list)  # type: ignore
    nullable: bool = True
    expectations: Optional[Expectation] = field(default_factory=list)
    identity: str = field(default_factory=random_string)
    precision: Optional[int] = None
    scale: Optional[int] = None
    origin: Optional[List[str]] = field(default_factory=list)

    def __init__(self, **kwargs):
        attributes = {f.name: f for f in fields(self.__class__)}
        for attribute in attributes:
            if attribute in kwargs:
                value = kwargs[attribute]
                # Special handling for 'expectations'
                if attribute == "expectations" and value:
                    value = [
                        (
                            v
                            if isinstance(v, Expectation)
                            else (
                                SchemaExpectation.load(v).update({"column": kwargs["name"]})
                                if v.get("column", _MISSING_VALUE) == _MISSING_VALUE
                                else v
                            )
                        )
                        for v in value
                    ]

                setattr(self, attribute, value)
            elif not isinstance(attributes[attribute].default, _MISSING_TYPE):
                setattr(self, attribute, attributes[attribute].default)
            elif not isinstance(attributes[attribute].default_factory, _MISSING_TYPE):
                setattr(self, attribute, attributes[attribute].default_factory())  # type:ignore
            else:
                raise ColumnDefinitionError(attribute)

        # map literals to OrsoTypes
        if self.type.__class__ is not OrsoTypes:
            type_name = str(self.type).upper()
            if type_name in OrsoTypes.__members__:
                self.type = OrsoTypes[type_name]
            elif type_name == "LIST":
                warn("Column type LIST will be deprecated in a future version, use ARRAY instead.")
                self.type = OrsoTypes.ARRAY
            elif type_name == "NUMERIC":
                warn(
                    "Column type NUMERIC will be deprecated in a future version, use DECIMAL, DOUBLE or INTEGER instead. Mapped to DOUBLE, this may not be compatible with all values NUMERIC was compatible with."
                )
                self.type = OrsoTypes.DOUBLE
            elif type_name == "STRING":
                raise ValueError(
                    f"Unknown column type '{self.type}' for column '{self.name}'. Did you mean 'VARCHAR'?"
                )
            elif self.type != 0:
                raise ValueError(f"Unknown column type '{self.type}' for column '{self.name}'.")

        if self.type == OrsoTypes.DECIMAL and self.precision is None:
            from decimal import getcontext

            self.precision = getcontext().prec
        if self.type == OrsoTypes.DECIMAL and self.scale is None:
            self.scale = int(0.75 * self.precision)

        # if we have a default value, parse it to the correct type and fail if we can't
        if self.default:
            try:
                self.default = self.type.parse(self.default)
            except Exception:
                raise ValueError(
                    f"Column '{self.name}' default value not compatible with '{self.type}'."
                )

    def __str__(self):
        return self.identity

    def materialize(self):
        raise TypeError("Cannot materialize FlatColumns")

    @classmethod
    def from_arrow(cls, arrow_field) -> "FlatColumn":
        """
        Converts a PyArrow field to an Orso FlatColumn object.

        Parameters:
            arrow_field: Field
                PyArrow Field object to be converted.

        Returns:
            FlatColumn: A FlatColumn object containing the converted information.
        """
        # Fetch the native type mapping from Arrow to Python native types
        native_type = arrow_type_map(arrow_field.type)
        # Initialize variables to hold optional decimal properties
        scale: Optional[int] = None
        precision: Optional[int] = None
        # Check if the type is Decimal and populate scale and precision
        if isinstance(native_type, Decimal):
            field_type = OrsoTypes.DECIMAL
            scale = native_type.scale  # type:ignore
            precision = native_type.precision  # type:ignore
        else:
            # Fall back to the generic mapping
            field_type = PYTHON_TO_ORSO_MAP.get(native_type, None)
            if field_type is None:
                raise ValueError(f"Unsupported type: {native_type}")

        return FlatColumn(
            name=str(arrow_field.name),
            type=field_type,
            nullable=arrow_field.nullable,
            scale=scale,
            precision=precision,
        )

    def to_flatcolumn(self) -> "FlatColumn":
        """
        Convert any column type to a FlatColumn (e.g. after a FunctionColumn has been evaluated)
        """
        return FlatColumn(
            name=str(self.name),
            default=self.default,
            description=self.description,
            aliases=self.aliases,
            identity=self.identity,
            type=self.type,
            nullable=self.nullable,
            scale=self.scale,
            precision=self.precision,
        )

    @property
    def all_names(self):
        """simplify collection of all of the names for this column"""
        if self.aliases is not None:
            return self.aliases + [self.name]
        return [self.name]

    @property
    def arrow_field(self):
        import pyarrow

        TYPE_MAP: dict = {
            OrsoTypes.BOOLEAN: pyarrow.bool_(),
            OrsoTypes.BLOB: pyarrow.binary(),
            OrsoTypes.DATE: pyarrow.date64(),
            OrsoTypes.TIMESTAMP: pyarrow.timestamp("us"),
            OrsoTypes.TIME: pyarrow.time32("ms"),
            OrsoTypes.INTERVAL: pyarrow.month_day_nano_interval(),
            OrsoTypes.STRUCT: pyarrow.struct([]),
            OrsoTypes.DECIMAL: lambda col: pyarrow.decimal128(col.precision, col.scale),
            OrsoTypes.DOUBLE: pyarrow.float64(),
            OrsoTypes.INTEGER: pyarrow.int64(),
            OrsoTypes.ARRAY: pyarrow.list_(pyarrow.string()),
            OrsoTypes.VARCHAR: pyarrow.string(),
            OrsoTypes.BSON: pyarrow.binary(),
            OrsoTypes.NULL: pyarrow.null(),
        }

        return pyarrow.field(name=self.name, type=TYPE_MAP.get(self.type, pyarrow.string()))

    def to_json(self) -> str:
        def default_serializer(o):
            if isinstance(o, OrsoTypes):
                return str(o)
            if isinstance(o, Expectation):
                return o.__dict__
            raise TypeError(f"Object of type {o.__class__.__name__} is not JSON serializable")

        return orjson.dumps(asdict(self), default=default_serializer)

    @classmethod
    def from_json(cls, json_str: str) -> "FlatColumn":
        def custom_decoder(dct):
            if "type" in dct:
                dct["type"] = OrsoTypes.__members__.get(dct["type"], OrsoTypes._MISSING_TYPE)
            if "expectations" in dct and isinstance(dct["expectations"], list):
                dct["expectations"] = [
                    SchemaExpectation.load(v) if isinstance(v, dict) else v
                    for v in dct["expectations"]
                ]
            return dct

        data = orjson.loads(json_str)
        return cls(**data)


@dataclass(init=False)
class FunctionColumn(FlatColumn):
    """
    This is a virtual column, it's nominally a column where the value is
    derived from a function.
    """

    binding: Optional[Callable] = lambda: None
    configuration: Tuple = field(default_factory=tuple)
    length: int = 1

    @property
    def values(self):
        raise TypeError("FunctionColumn has no 'values', do you mean 'materialize'?")

    def materialize(self):
        """
        Turn this virtual column into a list
        """
        value = self.binding(*self.configuration)
        return numpy.array([value] * self.length)


@dataclass(init=False)
class ConstantColumn(FlatColumn):
    """
    Rather than pass around columns of constant values, where we can we should
    replace them with this column type.

    Note: We don't implement anything here which deals with doing operations on
    two constant columns; whilst that would be a good optimization, the better
    way to do this is in the query optimizer, do operations on two constants
    while we're still working with a query plan.
    """

    length: int = 1
    value: Any = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.values = numpy.array([self.value])

    def materialize(self):
        """
        Turn this virtual column into a list.
        When performing element-wise operations, use value_array for broadcasting.
        """
        return numpy.full(self.length, self.values)


@dataclass(init=False)
class SparseColumn(FlatColumn):
    """
    This is a column type optimized for sparse data.
    Only the non-default values and their indices are stored.
    """

    values: numpy.ndarray = None
    default_value: Any = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        (self.indices,) = numpy.where(numpy.array(self.values) != self.default_value)
        self.values = numpy.array(self.values)[self.indices]
        self.total_length = len(kwargs.get("values", []))  # Store the total length

    def materialize(self):
        """
        Materialize the sparse column into a full numpy array.
        """
        materialized = numpy.full(
            self.total_length, self.default_value
        )  # Initialize with default values
        materialized[self.indices] = self.values
        return materialized


@dataclass(init=False)
class RLEColumn(FlatColumn):
    """
    This is a column type optimized for sequences of repeated values.
    """

    values: numpy.ndarray = None
    lengths: List[int] = field(default_factory=list)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        run_values = []
        run_lengths = []

        if len(self.values) == 0:
            self.values = numpy.array([])
            return

        prev_value = self.values[0]
        run_length = 1

        for value in self.values[1:]:
            if value == prev_value:
                run_length += 1
            else:
                run_values.append(prev_value)
                run_lengths.append(run_length)

                prev_value = value
                run_length = 1

        run_values.append(prev_value)
        run_lengths.append(run_length)

        self.values = numpy.array(run_values)
        self.lengths = run_lengths

    def materialize(self):
        """
        Turn this compressed column back into its original form.
        """
        materialized = []
        for value, length in zip(self.values, self.lengths):
            materialized.extend([value] * length)
        return numpy.array(materialized)


@dataclass(init=False)
class DictionaryColumn(FlatColumn):
    """
    If we know a column has a small amount of unique values AND is a large column
    AND we're going to perform an operation on the values, we should dictionary
    encode the column. This allows us to operate once on each unique value in the
    column, rather than each value in the column individually. At the cost of
    constructing and materializing the dictionary encoding.
    """

    values: List[Any] = field(default_factory=list)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        values = numpy.asarray(self.values)
        self.values, self.encoding = numpy.unique(values, return_inverse=True)

    def materialize(self):
        """
        Turn this virtual column into a list
        """
        return self.values[self.encoding]


@dataclass
class RelationSchema:
    name: str
    aliases: List[str] = field(default_factory=list)
    columns: List[FlatColumn] = field(default_factory=list)
    primary_key: str = None

    def __iter__(self):
        """Return an iterator over column names."""
        return iter([col.name for col in self.columns])

    def __add__(self, other: "RelationSchema") -> "RelationSchema":
        """
        When we add schemas together, we combine the set of columns.

        Parameters:
            other: RelationSchema
                The other schema to be added.

        Returns:
            RelationSchema: A new RelationSchema containing the combined columns.
        """
        new_schema = RelationSchema(name=self.name, aliases=self.aliases, columns=[])

        # Create a new list to hold the merged columns
        new_columns = self.columns[:]

        # Keep track of the seen identities - preload with the current set
        seen_identities = [col.identity for col in self.columns]

        for column in other.columns:
            if column.identity not in seen_identities:
                seen_identities.append(column.identity)
                new_columns.append(column)

        # Assign the new list of columns to the new schema
        new_schema.columns = new_columns

        return new_schema

    @property
    def num_columns(self) -> int:
        """Returns the number of columns in the schema."""
        return len(self.columns)

    def find_column(self, column_name: str) -> Optional[FlatColumn]:
        """
        Find a column by name or alias.

        Parameters:
            column_name: str
                Name or alias of the column.

        Returns:
            Optional[FlatColumn]: The FlatColumn object, if found. None otherwise.
        """
        for column in self.columns:
            if column_name in column.all_names:
                return column
        return None

    def all_column_names(self) -> List[str]:
        """
        Return all the names and aliases for columns in this relation.

        Returns:
            List[str]: List of all names and aliases.
        """

        def _inner():
            for column in self.columns:
                yield from column.all_names

        return list(_inner())

    @property
    def column_names(self) -> List[str]:
        """Return the names for columns in this relation."""
        return [column.name for column in self.columns]

    def column(self, i: Union[int, str]) -> Optional[FlatColumn]:
        """
        Get column by name or index.

        Parameters:
            i: Union[int, str]
                Index or name of the column.

        Returns:
            Optional[FlatColumn]: The FlatColumn object, if found. None otherwise.
        """
        if isinstance(i, int):
            return self.columns[i]
        else:
            return self.find_column(i)

    def pop_column(self, column_name: str) -> Optional[FlatColumn]:
        """
        Remove a column by its name and return it.

        Parameters:
            column_name: str
                Name of the column to be removed.

        Returns:
            Optional[FlatColumn]: The removed column if found, otherwise None.
        """
        for idx, column in enumerate(self.columns):
            if column.name == column_name:
                return self.columns.pop(idx)
        return None

    def to_dict(self) -> Dict:
        """
        Convert the Schema to a dictionary.

        Returns:
            Dict: A dictionary representation of the schema.
        """

        def _converter(obj) -> Dict:
            """Handle enum serialization."""
            return {key: value.value if isinstance(value, Enum) else value for key, value in obj}

        return asdict(self, dict_factory=_converter)

    @classmethod
    def from_dict(cls, dic: dict) -> "RelationSchema":
        """
        Create a Schema from a dictionary.

        Parameters:
            dic: Dict
                A dictionary to convert into a RelationSchema.

        Returns:
            RelationSchema: A new RelationSchema object.
        """
        schema = RelationSchema(name=dic["name"], aliases=dic.get("aliases", []))
        for column in dic["columns"]:
            if isinstance(column, dict):
                schema.columns.append(FlatColumn(**column))
            if isinstance(column, str):
                schema.columns.append(FlatColumn(name=column))
        return schema

    def validate(self, data: MutableMapping) -> bool:
        """
        Perform schema validation against a dictionary-formatted record.

        Parameters:
            data: MutableMapping
                A dictionary containing the data to validate against the schema.

        Returns:
            bool: True if the data is valid according to the schema.

        Raises:
            TypeError: If data is not dictionary-like.
            DataValidationError: If data validation fails.
        """
        if not isinstance(data, MutableMapping):
            raise TypeError("Cannot validate non Dictionary-type value")

        # Check if all fields in 'data' are in the schema
        extra_fields = set(data.keys()) - set(column.name for column in self.columns)
        if extra_fields:
            raise ExcessColumnsInDataError(columns=extra_fields)

        errors = defaultdict(list)

        for column in self.columns:
            if column.name not in data:
                errors["Column in Schema Not Found in Record"].append(column.name)

            else:
                value = data[column.name]

                if value is None:
                    if not column.nullable:
                        errors["Column not Nullable"].append(column.name)
                elif column.type != OrsoTypes._MISSING_TYPE and not isinstance(
                    value, ORSO_TO_PYTHON_MAP[column.type]
                ):
                    errors["Incorrect Type"].append(
                        (
                            column.name,
                            value,
                            column.type,
                        )
                    )

        if errors:
            raise DataValidationError(errors=errors)
        return True

    def to_json(self):
        return {
            "name": self.name,
            "aliases": self.aliases,
            "primary_key": self.primary_key,
            "columns": [col.to_json() for col in self.columns],
        }


def convert_arrow_schema_to_orso_schema(arrow_schema):
    return RelationSchema(
        name="arrow",
        columns=[FlatColumn.from_arrow(field) for field in arrow_schema],
    )


def convert_orso_schema_to_arrow_schema(orso_schema):
    from pyarrow import schema

    return schema([col.arrow_field for col in orso_schema.columns])



import os
import sys

import pyarrow
import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))

import orso
from orso.dataframe import DataFrame
from orso.exceptions import DataValidationError
from orso.row import Row
from tests import cities
from orso.schema import RelationSchema, FlatColumn
from orso.types import OrsoTypes


def create_schema():
    return RelationSchema(
        name="dataset",
        columns=[
            FlatColumn(name="A", type=OrsoTypes.INTEGER, nullable=False),
            FlatColumn(name="B", type=OrsoTypes.VARCHAR),
            FlatColumn(name="C", type=OrsoTypes.DOUBLE, nullable=False),
        ],
    )


def create_rows():
    row_factory = Row.create_class(schema=create_schema())
    return [
        row_factory([1, "a", 1.1]),
        row_factory([2, "b", 2.2]),
        row_factory([3, "c", 3.3]),
        row_factory([4, None, 4.4]),
        row_factory([5, "e", 5.5]),
    ]


def create_dataframe():
    schema = create_schema()
    rows = create_rows()
    return DataFrame(rows=rows, schema=schema)


def test_dataframe_materialize():
    dataframe = create_dataframe()
    dataframe.materialize()
    assert isinstance(dataframe._rows, list)


def test_dataframe_collect():
    dataframe = create_dataframe()
    result = dataframe.collect(["A", "C"])
    assert result == [[1, 2, 3, 4, 5], [1.1, 2.2, 3.3, 4.4, 5.5]], result

    result = dataframe.collect("A")
    assert result == [1, 2, 3, 4, 5], result


def test_dataframe_get_item():
    dataframe = create_dataframe()
    result = dataframe["A", "C"]
    assert result == [[1, 2, 3, 4, 5], [1.1, 2.2, 3.3, 4.4, 5.5]], result

    result = dataframe[0, 2]
    assert result == [[1, 2, 3, 4, 5], [1.1, 2.2, 3.3, 4.4, 5.5]], result

    result = dataframe["A", 2]
    assert result == [[1, 2, 3, 4, 5], [1.1, 2.2, 3.3, 4.4, 5.5]], result

    result = dataframe["A"]
    assert result == [1, 2, 3, 4, 5], result

    result = dataframe[0]
    assert result == [1, 2, 3, 4, 5], result


def test_dataframe_slice():
    dataframe = create_dataframe()
    result = dataframe.slice(offset=1, length=2)
    assert len(result) == 2


def test_dataframe_iter():
    dataframe = create_dataframe()
    assert len(list(dataframe)) == 5


def test_dataframe_len():
    dataframe = create_dataframe()
    assert len(dataframe) == 5


def test_dataframe_user_init():
    df = DataFrame(cities.values)
    assert df.column_names == ("name", "population", "country", "founded", "area", "language")
    assert df.rowcount == 20, df


def test_dataframe_head():
    df = DataFrame(cities.values)

    # Test default head() behavior (first 5 rows)
    head = df.head()
    assert head.rowcount == 5
    assert head.row(0)[0] == "Tokyo"

    # Test head() with size parameter
    head = df.head(3)
    assert head.rowcount == 3
    assert head.row(0)[0] == "Tokyo"


def test_dataframe_tail():
    df = DataFrame(cities.values)

    # Test default tail() behavior (last 5 rows)
    head = df.tail()
    assert head.rowcount == 5
    assert head.row(0)[0] == "Dubai", head.row(0)[0]

    # Test tail() with size parameter
    head = df.tail(3)
    assert head.rowcount == 3
    assert head.row(0)[0] == "Stockholm", head.row(0)[0]


def test_dataframe_filter():
    # Filter rows where column A is greater than 2
    dataframe = create_dataframe()
    mask = [row[0] > 2 for row in dataframe]
    filtered_dataframe = dataframe.filter(mask)
    assert len(filtered_dataframe) == 3
    assert filtered_dataframe.collect("A") == [3, 4, 5], filtered_dataframe.collect(["A"])


def test_take():
    # Create a DataFrame object with some data
    df = create_dataframe()

    # Test taking a subset of rows
    indexes = [0, 2, 4]
    expected_rows = [
        (1, "a", 1.1),
        (3, "c", 3.3),
        (5, "e", 5.5),
    ]
    result = df.take(indexes)
    result_rows = result.fetchall()
    assert result_rows == expected_rows, result_rows

    # Test taking all rows in the DataFrame
    indexes = [0, 1, 2, 3, 4]

    expected_rows = [
        (1, "a", 1.1),
        (2, "b", 2.2),
        (3, "c", 3.3),
        (4, None, 4.4),
        (5, "e", 5.5),
    ]
    result = df.take(indexes)
    result_rows = result.fetchall()
    assert result_rows == expected_rows, result_rows

    # Test taking an empty subset of rows
    indexes = []
    expected_rows = []
    result = df.take(indexes)
    assert result.fetchall() == expected_rows


def test_dataframe_hash():
    df1 = create_dataframe()
    df2 = create_dataframe()
    df3 = DataFrame(schema=create_schema(), rows=create_rows()[:4])

    assert hash(df1) == hash(df1)
    assert hash(df1) == hash(df2)
    assert hash(df1) != hash(df3)


def test_to_arrow():
    # Create a DataFrame
    df = create_dataframe()

    # Convert to PyArrow
    arrow_table = df.arrow()

    # Check the column names
    assert arrow_table.column_names == ["A", "B", "C"]

    # Check the number of rows
    assert arrow_table.num_rows == 5

    # Check the schema
    assert arrow_table.schema == pyarrow.schema(
        [("A", pyarrow.int64()), ("B", pyarrow.string()), ("C", pyarrow.float64())]
    )

    # Check the values
    expected_values = [(1, "a", 1.1), (2, "b", 2.2), (3, "c", 3.3), (4, None, 4.4), (5, "e", 5.5)]
    for i, col in enumerate(arrow_table.itercolumns()):
        assert col.to_pylist() == [v[i] for v in expected_values]


def test_to_arrow_with_size():
    df = create_dataframe()
    table = df.arrow(size=3)
    assert table.num_rows == 3, table

    df = create_dataframe()
    table = df.arrow(size=0)
    assert table.num_rows == 0, table


def test_appending():
    df = orso.DataFrame(schema=cities.schema)

    assert len(df) == 0
    # valid data item
    df.append(
        {
            "name": "Perth",
            "population": 2059484,
            "country": "Australia",
            "founded": "1829",
            "area": 6412.3,
            "language": "English",
        }
    )
    assert len(df) == 1

    # founded is nullable
    df.append(
        {
            "name": "Hobart",
            "population": 240342,
            "country": "Australia",
            "founded": None,  # actually 1804,
            "area": 1357.7,
            "language": "English",
        }
    )
    # country is not nullable
    with pytest.raises(DataValidationError):
        df.append(
            {
                "name": "Darwin",
                "population": 147255,
                "country": None,
                "founded": "1869",
                "area": 112.01,
                "language": "English",
            }
        )

    # population is an int
    with pytest.raises(DataValidationError):
        df.append(
            {
                "name": "Brisbane",
                "population": "2470394",  # is an int
                "country": "Australia",
                "founded": "1824",
                "area": 5905.9,
                "language": "English",
            }
        )


def test_profile():
    df = create_dataframe()
    profile = df.profile.to_dataframe()
    assert isinstance(profile, DataFrame)


def test_build_and_then_profile():
    df = orso.DataFrame(schema=cities.schema)
    for city in cities.values:
        df.append(city)

    p = df.profile.to_dataframe()
    assert p.rowcount == df.columncount
    assert p.collect("count") == [df.rowcount] * df.columncount


def test_describe():
    df = create_dataframe()
    desc = df.description
    assert len(desc) == df.columncount
    assert desc == [
        ("A", "INTEGER", None, None, None, None, False),
        ("B", "VARCHAR", None, None, None, None, True),
        ("C", "DOUBLE", None, None, None, None, False),
    ]


def test_distinct():
    df = orso.DataFrame(schema=cities.schema)
    for city in cities.values:
        df.append(city)
    for city in cities.values:
        df.append(city)

    assert df.distinct().rowcount == len(cities.values)


def test_to_map_and_to_dict():
    df = orso.DataFrame(cities.values)
    for row in df:
        assert isinstance(row, tuple)
        assert isinstance(row.as_map, tuple)
        assert isinstance(row.as_dict, dict)
        assert dict(row.as_map) == row.as_dict


def test_group_by():
    import opteryx

    gb = opteryx.query("SELECT * FROM $satellites").group_by("planetId").count()
    assert gb.shape == (7, 2)


def test_adding_dicts_in_wrong_order():
    df = orso.DataFrame(
        schema=RelationSchema(
            name="test",
            columns=[
                FlatColumn(name="column_1", type=OrsoTypes.INTEGER),
                FlatColumn(name="column_2", type=OrsoTypes.VARCHAR),
            ],
        )
    )

    # insert records with the dicts in different orders
    df.append({"column_1": 1, "column_2": "one"})
    df.append({"column_2": "two", "column_1": 2})

    # we should have them in the correct order when we extract them
    assert df.collect(["column_1", "column_2"]) == [[1, 2], ["one", "two"]], df.collect(
        ["column_1", "column_2"]
    )


def test_bytes():
    import opteryx
    import time

    gb = opteryx.query("SELECT * FROM $astronauts")
    t = time.monotonic_ns()
    first_bytes = gb.nbytes()
    first = time.monotonic_ns() - t
    t = time.monotonic_ns()
    second_bytes = gb.nbytes()
    second = time.monotonic_ns() - t

    assert first_bytes == second_bytes
    assert first > second


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



import os
import sys

import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))

import datetime
import pyarrow
import numpy

from orso.types import OrsoTypes

from orso.schema import FlatColumn
from orso.schema import FunctionColumn
from orso.schema import ConstantColumn
from orso.schema import DictionaryColumn
from orso.schema import SparseColumn
from orso.schema import RLEColumn

from orso.exceptions import ColumnDefinitionError


def test_flat_column_basic_init():
    column = FlatColumn(name="test", type=OrsoTypes.VARCHAR)
    assert column.name == "test"
    assert column.type == OrsoTypes.VARCHAR
    assert column.description is None
    assert column.aliases == []
    assert column.nullable is True


def test_flat_column_custom_values():
    column = FlatColumn(
        name="test",
        type="varchar",
        description="description",
        aliases=["alias1", "alias2"],
        nullable=False,
        precision=10,
        scale=5,
    )
    assert column.name == "test"
    assert column.type == OrsoTypes.VARCHAR
    assert column.description == "description"
    assert column.aliases == ["alias1", "alias2"]
    assert column.nullable is False
    assert column.precision == 10
    assert column.scale == 5


def test_flat_column_with_deprecated_type_list():
    with pytest.warns(UserWarning, match="Column type LIST will be deprecated"):
        column = FlatColumn(name="test", type="list")
    assert column.type == OrsoTypes.ARRAY


def test_flat_column_with_deprecated_type_numeric():
    with pytest.warns(UserWarning, match="Column type NUMERIC will be deprecated"):
        column = FlatColumn(name="test", type="numeric")
    assert column.type == OrsoTypes.DOUBLE


def test_flat_column_with_unknown_type():
    with pytest.raises(ValueError, match="Unknown column type"):
        FlatColumn(name="test", type="unknown")


def test_flat_column_materialize():
    flat_column = FlatColumn(name="athena", type=OrsoTypes.INTEGER)

    with pytest.raises(TypeError):
        flat_column.materialize()

    assert flat_column.name == "athena"
    assert flat_column.type == OrsoTypes.INTEGER


def test_flat_column_string_representation():
    column = FlatColumn(name="test", type=OrsoTypes.VARCHAR)
    assert str(column) == column.identity


def test_flat_column_missing_attribute():
    with pytest.raises(ColumnDefinitionError):
        FlatColumn()  # Missing required attributes like name and type


def test_columns_with_unknown_parameters():
    FlatColumn(name="athena", type=OrsoTypes.INTEGER, alpha="betty")
    FunctionColumn(name="aries", type=OrsoTypes.DATE, binding=datetime.date.today, sketty="yum")


def test_column_with_valid_default():
    col = FlatColumn(name="valid", type=OrsoTypes.INTEGER, default="1")
    assert col.default == 1

    col = FlatColumn(name="valid", type=OrsoTypes.INTEGER, default=1)
    assert col.default == 1


def test_column_with_invalid_default():
    with pytest.raises(ValueError):
        FlatColumn(name="invalid", type=OrsoTypes.INTEGER, default="green")


def test_flat_column_from_arrow():
    field_name = "test_field"
    arrow_type = pyarrow.string()
    nullable = True
    arrow_field = pyarrow.field(field_name, arrow_type, nullable)

    # Convert the arrow_field to a FlatColumn object
    column = FlatColumn.from_arrow(arrow_field)

    # Check that the name is correctly transferred
    assert column.name == arrow_field.name

    # Check that the type is correctly mapped
    assert column.type == OrsoTypes.VARCHAR, column.type

    # Check that the nullable property is correctly transferred
    assert column.nullable == arrow_field.nullable


def test_column_type_mapping():
    fc = FlatColumn(name="athena", type="INTEGER")
    assert fc.type == OrsoTypes.INTEGER
    assert fc.type.__class__ == OrsoTypes

    fc = FlatColumn(name="athens", type="LIST")
    assert fc.type == OrsoTypes.ARRAY, fc.type

    fc = FlatColumn(name="athens", type="NUMERIC")
    assert fc.type == OrsoTypes.DOUBLE, fc.type

    fc = FlatColumn(name="athled", type=0)
    assert fc.type == 0

    with pytest.raises(ValueError):
        FlatColumn(name="able", type="LEFT")


def test_missing_column_missing_name():
    with pytest.raises(ColumnDefinitionError):
        FlatColumn()


def test_type_checks():
    from decimal import Decimal
    from orso.schema import RelationSchema

    TEST_DATA = {
        OrsoTypes.VARCHAR: "string",
        OrsoTypes.INTEGER: 100,
        OrsoTypes.BOOLEAN: True,
        OrsoTypes.DATE: datetime.date.today(),
        OrsoTypes.ARRAY: ["a", "b", "c"],
        OrsoTypes.DOUBLE: 10.00,
        OrsoTypes.TIMESTAMP: datetime.datetime.utcnow(),
        OrsoTypes.TIME: datetime.time.min,
        OrsoTypes.BLOB: b"blob",
        OrsoTypes.DECIMAL: Decimal("3.7"),
        OrsoTypes.STRUCT: {"a": 1},
        OrsoTypes.INTERVAL: datetime.timedelta(days=1),
    }

    columns = []
    for t, v in TEST_DATA.items():
        columns.append(FlatColumn(name=str(t), type=t))

    schema = RelationSchema(name="test", columns=columns)
    schema.validate({str(k): v for k, v in TEST_DATA.items()})


def test_function_column():
    func_column = FunctionColumn(name="aries", type=OrsoTypes.DATE, binding=datetime.date.today)

    assert func_column.name == "aries"
    assert func_column.type == OrsoTypes.DATE
    assert func_column.length == 1
    assert func_column.aliases == []
    assert func_column.description == None

    values = func_column.materialize()

    assert len(values) == 1
    assert values[0] == datetime.date.today()

    func_column.length = 10
    values = func_column.materialize()

    assert len(values) == 10
    assert all(v == datetime.date.today() for v in values)


def test_constant_column():
    FACT: str = "The god of merchants, shepherds and messengers."

    const_column = ConstantColumn(name="hermes", type=OrsoTypes.VARCHAR, value=FACT)

    assert const_column.name == "hermes"
    assert const_column.type == OrsoTypes.VARCHAR
    assert const_column.value == FACT
    assert const_column.length == 1
    assert const_column.aliases == []
    assert const_column.description == None

    values = const_column.materialize()

    assert len(values) == 1
    assert values[0] == FACT

    const_column.length = 10
    values = const_column.materialize()

    assert len(values) == 10
    assert all(v == FACT for v in values)


def test_dict_column():
    MONTH_LENGTHS: list = ["31", "28", "31", "30", "31", "30", "31", "31", "30", "31", "30", "31"]

    dict_column = DictionaryColumn(name="pan", type=OrsoTypes.VARCHAR, values=MONTH_LENGTHS)

    assert dict_column.name == "pan"
    assert dict_column.type == OrsoTypes.VARCHAR

    assert sorted(dict_column.values) == ["28", "30", "31"]
    assert list(dict_column.encoding) == [2, 0, 2, 1, 2, 1, 2, 2, 1, 2, 1, 2]

    values = dict_column.materialize()

    assert list(values) == MONTH_LENGTHS


def test_rle_column():
    SEASON_LENGTHS: list = ["31", "30", "31", "31", "30", "31", "31", "31", "31"]

    rle_column = RLEColumn(name="season", type=OrsoTypes.VARCHAR, values=SEASON_LENGTHS)

    assert rle_column.name == "season"
    assert rle_column.type == OrsoTypes.VARCHAR

    assert numpy.array_equal(rle_column.values, ["31", "30", "31", "30", "31"])
    assert rle_column.lengths == [1, 1, 2, 1, 4]

    values = rle_column.materialize()

    assert list(values) == SEASON_LENGTHS


def test_sparse_column():
    VARYING_LENGTHS: list = ["31", None, "31", None, None, "31", "30", "31", None]

    sparse_column = SparseColumn(name="varying", type=OrsoTypes.VARCHAR, values=VARYING_LENGTHS)

    assert sparse_column.name == "varying"
    assert sparse_column.type == OrsoTypes.VARCHAR

    assert numpy.array_equal(sparse_column.values, numpy.array(["31", "31", "31", "30", "31"]))
    assert numpy.array_equal(sparse_column.indices, numpy.array([0, 2, 5, 6, 7]))

    values = sparse_column.materialize()

    assert numpy.array_equal(values, numpy.array(VARYING_LENGTHS))


import numpy


def test_sparse_column_multiply():
    # Initialize the sparse column
    original_values = [1, None, 2, None, None, 3, 4, 5, None]
    sparse_col = SparseColumn(name="test", type=OrsoTypes.INTEGER, values=original_values)

    # Perform the operation on compressed values
    sparse_col.values = sparse_col.values * 2

    # Materialize and compare
    materialized_values = sparse_col.materialize()
    expected_values = [v * 2 if v is not None else None for v in original_values]

    # Convert to numpy array for direct comparison
    expected_values_np = numpy.array(expected_values, dtype=object)

    numpy.testing.assert_array_equal(materialized_values, expected_values_np)


# Constant Column Test
def test_constant_column_multiply():
    constant_col = ConstantColumn(name="const", type=OrsoTypes.INTEGER, length=5, value=3)
    constant_col.values *= 2
    materialized_values = constant_col.materialize()
    expected_values_np = numpy.full((5,), 6)
    numpy.testing.assert_array_equal(materialized_values, expected_values_np)


# Dictionary Column Test
def test_dictionary_column_multiply():
    original_values = [1, 3, 2, 2, 3, 1]
    dict_col = DictionaryColumn(name="dict", type=OrsoTypes.INTEGER, values=original_values)
    dict_col.values = dict_col.values * 2
    materialized_values = dict_col.materialize()
    expected_values = [v * 2 for v in original_values]
    expected_values_np = numpy.array(expected_values)
    numpy.testing.assert_array_equal(materialized_values, expected_values_np)


# RLE Column Test
def test_rle_column_multiply():
    original_values = [1, 1, 2, 2, 3, 3]
    rle_col = RLEColumn(name="rle", type=OrsoTypes.INTEGER, values=original_values)
    rle_col.values *= 2
    materialized_values = rle_col.materialize()
    expected_values = [v * 2 for v in original_values]
    expected_values_np = numpy.array(expected_values)
    numpy.testing.assert_array_equal(materialized_values, expected_values_np)


def test_to_flatcolumn_basic():
    """
    Test that to_flatcolumn returns a new FlatColumn object.
    """
    flat_column = FlatColumn(
        name="id",
        type=OrsoTypes.INTEGER,
        description="An ID column",
        aliases=["ID"],
        nullable=False,
        precision=5,
    )

    new_column = flat_column.to_flatcolumn()
    assert isinstance(new_column, FlatColumn)
    assert new_column is not flat_column


def test_to_flatcolumn_from_function_column():
    """
    Test that to_flatcolumn returns a new FlatColumn object.
    """
    func_column = FunctionColumn(
        name="virtual_id", type=OrsoTypes.INTEGER, binding=lambda: 10, length=10
    )

    new_column = func_column.to_flatcolumn()
    assert isinstance(new_column, FlatColumn)
    assert new_column is not func_column


def test_to_flatcolumn_preserve_attributes():
    """
    Test that to_flatcolumn preserves the attributes.
    """
    flat_column = FlatColumn(
        name="id",
        type=OrsoTypes.INTEGER,
        description="An ID column",
        aliases=["ID"],
        nullable=False,
        precision=5,
    )

    new_column = flat_column.to_flatcolumn()

    for field in [
        f
        for f in dir(FlatColumn)
        if (f[0] != "_" and isinstance(getattr(flat_column, f), (int, str, float, list, OrsoTypes)))
    ]:
        assert getattr(new_column, field) == getattr(flat_column, field), field


def test_to_json_and_back():
    """
    Test that to_flatcolumn preserves the attributes.
    """
    flat_column = FlatColumn(
        name="id",
        type=OrsoTypes.INTEGER,
        description="An ID column",
        aliases=["ID"],
        nullable=False,
        precision=5,
    )
    as_json = flat_column.to_json()
    as_column = FlatColumn.from_json(as_json)

    for field in [
        f
        for f in dir(FlatColumn)
        if (f[0] != "_" and isinstance(getattr(flat_column, f), (int, str, float, list, OrsoTypes)))
    ]:
        assert getattr(as_column, field) == getattr(flat_column, field), field


def test_aliasing():
    col = FlatColumn(name="alpha", type=OrsoTypes.VARCHAR)

    assert col.all_names == ["alpha"]

    col.aliases = ["beta"]
    assert set(col.all_names) == {"alpha", "beta"}

    col.aliases = ["gamma", "delta"]
    assert set(col.all_names) == {"alpha", "gamma", "delta"}


def test_minimum_definition():
    col = FlatColumn(name="a")


def test_arrow_conversion():
    from tests.cities import schema as city_schema
    from pyarrow import schema as arrow_schema

    _arrow_schema = arrow_schema([col.arrow_field for col in city_schema.columns])

    print(_arrow_schema)


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.types import OrsoTypes


def test_types_is_numeric():
    # is numeric
    assert not OrsoTypes.ARRAY.is_numeric()
    assert not OrsoTypes.BLOB.is_numeric()
    assert not OrsoTypes.BOOLEAN.is_numeric()
    assert not OrsoTypes.DATE.is_numeric()
    assert OrsoTypes.DECIMAL.is_numeric()
    assert OrsoTypes.DOUBLE.is_numeric()
    assert OrsoTypes.INTEGER.is_numeric()
    assert not OrsoTypes.INTERVAL.is_numeric()
    assert not OrsoTypes.STRUCT.is_numeric()
    assert not OrsoTypes.TIME.is_numeric()
    assert not OrsoTypes.TIMESTAMP.is_numeric()
    assert not OrsoTypes.VARCHAR.is_numeric()


def test_types_is_temporal():
    # is temporal
    assert not OrsoTypes.ARRAY.is_temporal()
    assert not OrsoTypes.BLOB.is_temporal()
    assert not OrsoTypes.BOOLEAN.is_temporal()
    assert OrsoTypes.DATE.is_temporal()
    assert not OrsoTypes.DECIMAL.is_temporal()
    assert not OrsoTypes.DOUBLE.is_temporal()
    assert not OrsoTypes.INTEGER.is_temporal()
    assert not OrsoTypes.INTERVAL.is_temporal()
    assert not OrsoTypes.STRUCT.is_temporal()
    assert OrsoTypes.TIME.is_temporal()
    assert OrsoTypes.TIMESTAMP.is_temporal()
    assert not OrsoTypes.VARCHAR.is_temporal()


def test_types_is_large_object():
    # is temporal
    assert not OrsoTypes.ARRAY.is_large_object()
    assert OrsoTypes.BLOB.is_large_object()
    assert not OrsoTypes.BOOLEAN.is_large_object()
    assert not OrsoTypes.DATE.is_large_object()
    assert not OrsoTypes.DECIMAL.is_large_object()
    assert not OrsoTypes.DOUBLE.is_large_object()
    assert not OrsoTypes.INTEGER.is_large_object()
    assert not OrsoTypes.INTERVAL.is_large_object()
    assert not OrsoTypes.STRUCT.is_large_object()
    assert not OrsoTypes.TIME.is_large_object()
    assert not OrsoTypes.TIMESTAMP.is_large_object()
    assert OrsoTypes.VARCHAR.is_large_object()


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.compute.bloom_filter import create_bloom_filter

ITERATIONS: int = 1000


def random_string(width: int = 16):
    import random

    num_chars = ((width + 1) >> 1) << 3  # Convert length to number of bits
    rand_bytes = random.getrandbits(num_chars)  # Generate random bytes
    rand_hex = hex(rand_bytes)[
        2 : width + 2
    ]  # Convert to hex string and truncate to desired length
    return rand_hex


def test_bloom_filter():
    # first we populate the BloomFilter
    tokens = (random_string(48) for i in range(ITERATIONS))
    bf = create_bloom_filter(19000, [])
    for token in tokens:
        bf.add(hash(token))

    # then we test. 100% shouldn't match (we use different string lengths)
    # but we're a probabilistic filter so expect some false positives
    tokens = (hash(random_string(32)) for i in range(ITERATIONS))
    collisions = 0
    for token in tokens:
        if bf.possibly_contains(hash(token)):
            collisions += 1

    # this is approximately 1% false positive rate, we're going to test between
    # 0.5% and 5% because this is probabilistic so are unlikely to actually get 1%
    assert (ITERATIONS * 0.005) < collisions < (ITERATIONS * 0.05), collisions / (ITERATIONS)


def test_bloom_contains():
    bf = create_bloom_filter(1000, [123])
    assert bf.possibly_contains(123)
    assert not bf.possibly_contains(888) and not bf.possibly_contains(321)


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



# cython: language_level=3

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module was written with assistance from ChatGPT
"""

cdef extern from "Python.h":
    void *PyMem_Malloc(int size) nogil
    void PyMem_Free(void *ptr) nogil
    void memset(void *s, int c, int n)

cdef class BitArray:
    cdef public int size
    cdef long long *bits

    def __init__(self, int size):
        assert size > 0, "bitarray size must be a positive integer"
        self.size = size
        cdef int n_longs = (size + 63) >> 6
        self.bits = <long long *> PyMem_Malloc(n_longs * sizeof(long long))
        memset(self.bits, 0, n_longs * sizeof(long long))

    def __dealloc__(self):
        with nogil:
            PyMem_Free(self.bits)

    def get(self, int index):
        if 0 > index or index >= self.size:
            raise IndexError("Index out of range")
        return (self.bits[index >> 6] & (1 << (index & 63))) != 0

    def set(self, int index, bint value):
        if 0 > index or index >= self.size:
            raise IndexError("Index out of range")
        if value:
            self.bits[index >> 6] |= (1 << (index & 63))
        else:
            self.bits[index >> 6] &= ~(1 << (index & 63))

    @property
    def array(self):
        ba = bytearray((self.size + 7) >> 3)
        for i in range((self.size + 63) >> 6):
            for j in range(64):
                if i * 64 + j < self.size:
                    ba[i*8 + (j>>3)] |= ((self.bits[i] >> j) & 1) << (j & 7)
        return ba

    @classmethod
    def from_array(cls, array, int length):
        bit_array = cls(length)
        for index in range(length):
            bit = (array[(index >> 3)] & (1 << (index & 7))) != 0
            if index < length and bit:
                bit_array.set(index, 1)
        return bit_array



import os
import sys
import pytest

sys.path.insert(1, os.path.join(sys.path[0], ".."))

import pytest
from unittest.mock import Mock, patch
from orso.tools import retry  # assuming the decorator is saved in retry_decorator.py


def test_retry_success_on_first_try():
    mock_func = Mock(return_value="success")

    @retry(max_tries=3)
    def decorated_func():
        return mock_func()

    with patch("time.sleep", return_value=None) as mock_sleep:
        result = decorated_func()
        assert result == "success"
        mock_func.assert_called_once()
        mock_sleep.assert_not_called()


def test_retry_success_after_retries():
    mock_func = Mock(side_effect=[Exception("fail"), Exception("fail"), "success"])

    @retry(max_tries=3, backoff_seconds=1)
    def decorated_func():
        return mock_func()

    with patch("time.sleep", return_value=None) as mock_sleep:
        result = decorated_func()
        assert result == "success"
        assert mock_func.call_count == 3
        assert mock_sleep.call_count == 2


def test_retry_failure_after_max_retries():
    mock_func = Mock(side_effect=Exception("fail"))

    @retry(max_tries=3, backoff_seconds=1)
    def decorated_func():
        return mock_func()

    with patch("time.sleep", return_value=None) as mock_sleep:
        with pytest.raises(Exception):
            decorated_func()

        assert mock_func.call_count == 3
        assert mock_sleep.call_count == 2


def test_retry_with_exponential_backoff():
    mock_func = Mock(side_effect=[Exception("fail"), Exception("fail"), "success"])

    @retry(max_tries=3, backoff_seconds=1, exponential_backoff=True, max_backoff=4)
    def decorated_func():
        return mock_func()

    with patch("time.sleep", return_value=None) as mock_sleep:
        result = decorated_func()
        assert result == "success"
        assert mock_func.call_count == 3
        assert mock_sleep.call_count == 2
        mock_sleep.assert_any_call(1)
        mock_sleep.assert_any_call(2)


def test_retry_with_jitter():
    mock_func = Mock(side_effect=[Exception("fail"), Exception("fail"), "success"])

    @retry(max_tries=3, backoff_seconds=1, jitter=True)
    def decorated_func():
        return mock_func()

    with (
        patch("time.sleep", return_value=None) as mock_sleep,
        patch("random.uniform", return_value=0.5) as mock_random,
    ):
        result = decorated_func()
        assert result == "success"
        assert mock_func.call_count == 3
        assert mock_sleep.call_count == 2
        mock_random.assert_called()


def test_retry_with_specific_exceptions():
    mock_func = Mock(side_effect=[ValueError("fail"), ValueError("fail"), "success"])

    @retry(max_tries=3, backoff_seconds=1, retry_exceptions=(ValueError,))
    def decorated_func():
        return mock_func()

    with patch("time.sleep", return_value=None) as mock_sleep:
        result = decorated_func()
        assert result == "success"
        assert mock_func.call_count == 3
        assert mock_sleep.call_count == 2


def test_retry_with_callback():
    mock_func = Mock(side_effect=[Exception("fail"), "success"])
    callback_func = Mock()

    @retry(max_tries=3, backoff_seconds=1, callback=callback_func)
    def decorated_func():
        return mock_func()

    with patch("time.sleep", return_value=None) as mock_sleep:
        result = decorated_func()
        assert result == "success"
        assert mock_func.call_count == 2
        assert mock_sleep.call_count == 1

        # Ensure callback was called once with correct arguments
        assert callback_func.call_count == 1
        called_exception, called_attempt = callback_func.call_args[0]
        assert isinstance(called_exception, Exception)
        assert called_exception.args == ("fail",)
        assert called_attempt == 1


def test_retry_no_retries_needed():
    mock_func = Mock(return_value="success")

    @retry(max_tries=3, backoff_seconds=1)
    def decorated_func():
        return mock_func()

    with patch("time.sleep", return_value=None) as mock_sleep:
        result = decorated_func()
        assert result == "success"
        assert mock_func.call_count == 1
        mock_sleep.assert_not_called()


def test_retry_with_multiple_exceptions():
    mock_func = Mock(side_effect=[ValueError("fail"), KeyError("fail"), "success"])

    @retry(max_tries=3, backoff_seconds=1, retry_exceptions=(ValueError, KeyError))
    def decorated_func():
        return mock_func()

    with patch("time.sleep", return_value=None) as mock_sleep:
        result = decorated_func()
        assert result == "success"
        assert mock_func.call_count == 3
        assert mock_sleep.call_count == 2


def test_retry_with_partial_success():
    mock_func = Mock(
        side_effect=[Exception("fail"), "partial_success", Exception("fail_again"), "success"]
    )

    @retry(max_tries=4, backoff_seconds=1)
    def decorated_func():
        result = mock_func()
        if result == "partial_success":
            raise Exception("retrying due to partial success")
        return result

    with patch("time.sleep", return_value=None) as mock_sleep:
        result = decorated_func()
        assert result == "success"
        assert mock_func.call_count == 4
        assert mock_sleep.call_count == 3


def test_retry_with_max_backoff():
    mock_func = Mock(side_effect=[Exception("fail"), Exception("fail"), "success"])

    @retry(max_tries=3, backoff_seconds=1, exponential_backoff=True, max_backoff=2)
    def decorated_func():
        return mock_func()

    with patch("time.sleep", return_value=None) as mock_sleep:
        result = decorated_func()
        assert result == "success"
        assert mock_func.call_count == 3
        assert mock_sleep.call_count == 2
        mock_sleep.assert_any_call(1)
        mock_sleep.assert_any_call(2)


if __name__ == "__main__":  # pragma: nocover
    from tests import run_tests

    run_tests()



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.dataframe import DataFrame


def test_fetchall():
    dataframe = DataFrame(
        rows=[(1, "John"), (2, "Jane"), (3, "Bob")],
        schema=["id", "name"],
    )
    result = dataframe.fetchall()
    expected = [(1, "John"), (2, "Jane"), (3, "Bob")]
    assert result == expected
    assert dataframe.fetchall() == [], dataframe.fetchall()


def test_fetchone():
    dataframe = DataFrame(
        rows=[(1, "John"), (2, "Jane"), (3, "Bob")],
        schema=["id", "name"],
    )
    result1 = dataframe.fetchone()
    expected1 = (1, "John")
    assert result1 == expected1

    result2 = dataframe.fetchone()
    expected2 = (2, "Jane")
    assert result2 == expected2

    result3 = dataframe.fetchone()
    expected3 = (3, "Bob")
    assert result3 == expected3

    result4 = dataframe.fetchone()
    expected4 = None
    assert result4 == expected4


def test_fetchmany():
    dataframe = DataFrame(
        rows=[(1, "John"), (2, "Jane"), (3, "Bob")],
        schema=["id", "name"],
    )
    result1 = dataframe.fetchmany(2)
    expected1 = [(1, "John"), (2, "Jane")]
    assert result1 == expected1, result1

    result2 = dataframe.fetchmany(2)
    expected2 = [(3, "Bob")]
    assert result2 == expected2, result2

    result3 = dataframe.fetchmany(2)
    expected3 = []
    assert result3 == expected3, result3


def test_fetch_methods():
    dataframe = DataFrame(
        rows=[(1, "John"), (2, "Jane"), (3, "Bob")],
        schema=["id", "name"],
    )

    # Test fetchone
    result1 = dataframe.fetchone()
    expected1 = (1, "John")
    assert result1 == expected1

    # Test fetchmany
    result2 = dataframe.fetchmany(2)
    expected2 = [(2, "Jane"), (3, "Bob")]
    assert result2 == expected2

    # Test fetchall
    result3 = dataframe.fetchall()
    expected3 = []
    assert result3 == expected3


if __name__ == "__main__":  # prgama: nocover
    from tests import run_tests

    run_tests()



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
"""
approximately 500 items, with two hashes, in 4092 bits, would have a 5% FP rate.
This implementation runs in about 3x faster than previous implementations.

This is not a general perpose Bloom Filter, if used outside of Orso, it may not
perform entirely as expected.
"""


from libc.stdlib cimport malloc, free
from libc.string cimport memset, memcpy

cdef class BloomFilter:
    cdef:
        unsigned char* bit_array
        long size

    def __cinit__(self, long size):
        self.size = size
        # Allocate memory for the bit array and initialize to 0
        self.bit_array = <unsigned char*>malloc(size // 8 + 1)
        if not self.bit_array:
            raise MemoryError("Failed to allocate memory for the bit array.")
        memset(self.bit_array, 0, size // 8 + 1)

    def __dealloc__(self):
        if self.bit_array:
            free(self.bit_array)

    cpdef void add(self, long item):
        """Add an item to the Bloom filter"""
        h1 = item % self.size
        # Apply the golden ratio to the item and use modulo to wrap within the size of the bit array
        h2 = <long>(item * 1.618033988749895) % self.size
        # Set bits using bitwise OR
        self.bit_array[h1 // 8] |= 1 << (h1 % 8)
        self.bit_array[h2 // 8] |= 1 << (h2 % 8)

    cpdef int possibly_contains(self, long item):
        """Check if the item might be in the set"""
        h1 = item % self.size
        # Apply the golden ratio to the item and use modulo to wrap within the size of the bit array
        h2 = <long>(item * 1.618033988749895) % self.size
        # Check bits using bitwise AND
        return (self.bit_array[h1 // 8] & (1 << (h1 % 8))) and \
               (self.bit_array[h2 // 8] & (1 << (h2 % 8)))

    cpdef memoryview serialize(self):
        """Serialize the Bloom filter to a memory view"""
        return memoryview(self.bit_array)[:self.size // 8 + 1]

    @staticmethod
    cdef BloomFilter deserialize(const unsigned char* data, long size):
        """Deserialize a memory view to a Bloom filter"""
        bf = BloomFilter(size)
        memcpy(bf.bit_array, data, size // 8 + 1)
        return bf

def create_bloom_filter(int size, items):
    """Create and populate a Bloom filter"""
    cdef BloomFilter bf = BloomFilter(size)
    for item in items:
        bf.add(item)
    return bf



import decimal
import random


def generate_random_decimal(precision: int, scale: int) -> decimal.Decimal:
    """
    Generates a random decimal number based on the given precision and scale.

    Parameters:
        precision: int
            The total number of digits in the decimal number.

        scale: int
            The number of digits to the right of the decimal point.

    Returns:
        Decimal: Randomly generated decimal number.
    """
    if scale > precision:
        raise ValueError("Scale can't be greater than precision")

    int_part_len = precision - scale

    int_part = "".join(str(random.randint(0, 9)) for _ in range(int_part_len))
    frac_part = "".join(str(random.randint(0, 9)) for _ in range(scale))

    # Ensure the integer part is not all zeros.
    if all(d == "0" for d in int_part):  # pragma: no cover
        int_part = str(random.randint(1, 9)) + int_part[1:]

    return decimal.Decimal(f"{int_part}.{frac_part}")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MissingDependencyError(Exception):
    def __init__(self, dependency):
        self.dependency = dependency
        message = f"No module named '{dependency}' can be found, "
        "please install or include in requirements.txt"
        super().__init__(message)


class DataError(Exception):
    pass


class DataValidationError(DataError):
    def __init__(self, errors: dict):
        truncated_text = lambda text: (
            f"{text[:16]}{'...' if len(text) > 16 else ''}... [{len(text) - 16} more]"
            if len(text) > 16
            else text
        )

        self.errors = errors

        message = "Data did not pass validation checks; "
        for err, detail in errors.items():
            message += f"\n{err}: "
            if all(isinstance(e, str) for e in detail):
                message += ", ".join(f"`{d}`" for d in detail)
            else:
                message += ", ".join(
                    [
                        f"`{d[0]}` value of `{truncated_text(str(d[1]))}` is not a {d[2].name}"
                        for d in detail
                    ]
                )

        super().__init__(message)


class ExcessColumnsInDataError(DataError):
    def __init__(self, columns):
        self.columns = columns
        message = (
            f"Data did not pass validation checks; "
            f"Additional fields, not defined in the schema, were present in the record - "
            + ", ".join(columns)
        )
        super().__init__(message)


class ColumnDefinitionError(DataError):
    def __init__(self, attribute):
        self.attribute = attribute
        message = f"Column is missing attribute {attribute}"
        super().__init__(message)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import decimal
from enum import Enum
from typing import Any

from orso.tools import parse_iso


class OrsoTypes(str, Enum):
    """
    The names of the types supported by Orso
    """

    ARRAY = "ARRAY"
    BLOB = "BLOB"
    BOOLEAN = "BOOLEAN"
    DATE = "DATE"
    DECIMAL = "DECIMAL"
    DOUBLE = "DOUBLE"
    INTEGER = "INTEGER"
    INTERVAL = "INTERVAL"
    STRUCT = "STRUCT"
    TIMESTAMP = "TIMESTAMP"
    TIME = "TIME"
    VARCHAR = "VARCHAR"
    NULL = "NULL"
    BSON = "BSON"
    _MISSING_TYPE = 0

    def is_numeric(self):
        """is the typle number-based"""
        return self in (self.INTEGER, self.DOUBLE, self.DECIMAL)

    def is_temporal(self):
        """is the type time-based"""
        return self in (self.DATE, self.TIME, self.TIMESTAMP)

    def is_large_object(self):
        """is the type arbitrary length string"""
        return self in (self.VARCHAR, self.BLOB)

    def is_complex(self):
        return self in (self.ARRAY, self.STRUCT, self.BSON, self.INTERVAL)

    def __str__(self):
        return self.value

    def parse(self, value: Any) -> Any:
        return ORSO_TO_PYTHON_PARSER[self.value](value)


ORSO_TO_PYTHON_MAP: dict = {
    OrsoTypes.BOOLEAN: bool,
    OrsoTypes.BLOB: bytes,
    OrsoTypes.DATE: datetime.date,
    OrsoTypes.TIMESTAMP: datetime.datetime,
    OrsoTypes.TIME: datetime.time,
    OrsoTypes.INTERVAL: datetime.timedelta,
    OrsoTypes.STRUCT: dict,
    OrsoTypes.DECIMAL: decimal.Decimal,
    OrsoTypes.DOUBLE: float,
    OrsoTypes.INTEGER: int,
    OrsoTypes.ARRAY: list,
    OrsoTypes.VARCHAR: str,
    OrsoTypes.BSON: bytes,
    OrsoTypes.NULL: None,
}

PYTHON_TO_ORSO_MAP: dict = {
    value: key for key, value in ORSO_TO_PYTHON_MAP.items() if key != OrsoTypes.BSON
}
PYTHON_TO_ORSO_MAP.update({tuple: OrsoTypes.ARRAY, set: OrsoTypes.ARRAY})  # map other python types

ORSO_TO_PYTHON_PARSER: dict = {
    OrsoTypes.BOOLEAN: bool,
    OrsoTypes.BLOB: bytes,
    OrsoTypes.DATE: lambda x: parse_iso(x).date(),
    OrsoTypes.TIMESTAMP: parse_iso,
    OrsoTypes.TIME: lambda x: parse_iso(x).time(),
    OrsoTypes.INTERVAL: datetime.timedelta,
    OrsoTypes.STRUCT: dict,
    OrsoTypes.DECIMAL: decimal.Decimal,
    OrsoTypes.DOUBLE: float,
    OrsoTypes.INTEGER: int,
    OrsoTypes.ARRAY: list,
    OrsoTypes.VARCHAR: str,
    OrsoTypes.BSON: bytes,
    OrsoTypes.NULL: lambda x: None,
}
